"""Types for MADSci Workcell configuration."""

from pathlib import Path
from typing import Annotated, ClassVar, Literal, Optional, Union

from madsci.common.serializers import dict_to_list
from madsci.common.types.base_types import (
    BaseModel,
    LoadConfig,
    ModelLink,
    PathLike,
    new_ulid_str,
)
from madsci.common.types.event_types import EventClientConfig
from madsci.common.types.lab_types import ManagerType
from madsci.common.types.location_types import Location
from madsci.common.types.node_types import NodeDefinition
from madsci.common.validators import create_dict_promoter, ulid_validator
from pydantic import computed_field, field_serializer
from pydantic.functional_validators import field_validator
from pydantic.networks import AnyUrl
from sqlmodel.main import Field


class WorkcellDefinition(BaseModel, extra="allow"):
    """Configuration for a MADSci Workcell."""

    _definition_file_patterns: ClassVar[list] = [
        "*workcell.yaml",
        "*workcell.yml",
        "*workcell.manager.yml",
        "*workcell.manager.yaml",
    ]
    _definition_cli_flag: ClassVar[list] = [
        "--workcell",
        "--workcell-definition",
        "--definition",
        "--workcell-definition-file",
        "-f",
    ]

    workcell_name: str = Field(
        title="Workcell Name", description="The name of the workcell."
    )
    manager_type: Literal[ManagerType.WORKCELL_MANAGER] = Field(
        title="Manager Type",
        description="The type of manager",
        default=ManagerType.WORKCELL_MANAGER,
    )
    workcell_id: str = Field(
        title="Workcell ID",
        description="The ID of the workcell.",
        default_factory=new_ulid_str,
    )
    description: Optional[str] = Field(
        default=None,
        title="Workcell Description",
        description="A description of the workcell.",
    )
    config: Annotated[
        "WorkcellConfig",
        Field(
            title="Workcell Configuration",
            description="The configuration for the workcell.",
            default_factory=lambda: WorkcellConfig(),
        ),
        LoadConfig(use_fields_as_cli_args=True),
    ]
    nodes: dict[str, Union["NodeDefinition", AnyUrl, PathLike]] = Field(
        default_factory=dict,
        title="Workcell Node URLs",
        description="The URL, path, or definition for each node in the workcell.",
    )
    locations: list[Location] = Field(
        default_factory=list,
        title="Workcell Locations",
        description="The Locations used in the workcell.",
    )

    @computed_field
    @property
    def workcell_directory(self) -> Path:
        """The directory for the workcell."""
        return Path(self.config.workcells_directory) / self.workcell_name

    is_ulid = field_validator("workcell_id")(ulid_validator)
    validate_nodes_to_dict = field_validator("nodes", mode="before")(
        create_dict_promoter("node_name")
    )
    serialize_nodes_to_list = field_serializer("nodes")(dict_to_list)


class WorkcellLink(ModelLink[WorkcellDefinition]):
    """Link to a MADSci Workcell Definition."""

    definition: Optional[WorkcellDefinition] = Field(
        title="Workcell Definition",
        description="The actual definition of the workcell.",
        default=None,
    )


class WorkcellConfig(BaseModel):
    """Configuration for a MADSci Workcell."""

    host: str = Field(
        default="127.0.0.1",
        title="Host",
        description="The host to run the workcell manager on.",
    )
    port: int = Field(
        default=8013,
        title="Port",
        description="The port to run the workcell manager on.",
    )
    workcells_directory: Optional[PathLike] = Field(
        title="Workcells Directory",
        description="Directory used to store workcell-related files in. Defaults to ~/.madsci/workcells. Workcell-related filess will be stored in a sub-folder with the workcell name.",
        default_factory=lambda: Path("~") / ".madsci" / "workcells",
    )
    redis_host: str = Field(
        default="localhost",
        title="Redis Host",
        description="The hostname for the redis server .",
    )
    redis_port: int = Field(
        default=6379,
        title="Redis Port",
        description="The port for the redis server.",
    )
    redis_password: Union[str, None] = Field(
        default=None,
        title="Redis Password",
        description="The password for the redis server.",
    )
    event_client_config: Optional[EventClientConfig] = Field(
        default=None,
        title="Event Client Configuration",
        description="The configuration for a MADSci event client.",
    )
    scheduler_update_interval: float = Field(
        default=2.0,
        title="Scheduler Update Interval",
        description="The interval at which the scheduler runs, in seconds. Must be >= node_update_interval",
    )
    node_update_interval: float = Field(
        default=1.0,
        title="Node Update Interval",
        description="The interval at which the workcell queries its node's states, in seconds.Must be <= scheduler_update_interval",
    )
    heartbeat_interval: float = Field(
        default=2.0,
        title="Heartbeat Interval",
        description="The interval at which the workcell queries its node's states, in seconds.Must be <= scheduler_update_interval",
    )
    auto_start: bool = Field(
        default=True,
        title="Auto Start",
        description="Whether the workcell should automatically create a new Workcell Manager and start it when the lab starts, registering it with the Lab Server.",
    )
    clear_workflows: bool = Field(
        default=False,
        title="Clear Workflows",
        description="Whether the workcell should clear old workflows on restart",
    )
    cold_start_delay: int = Field(
        default=0,
        title="Cold Start Delay",
        description="How long the Workcell engine should sleep on startup",
    )
    scheduler: str = Field(
        default="madsci.workcell_manager.schedulers.default_scheduler",
        title="scheduler",
        description="Scheduler module that contains a Scheduler class that inherits from AbstractScheduler to use",
    )
    data_server_url: Optional[AnyUrl] = Field(
        default=None,
        title="Data Client URL",
        description="The URL for the data client.",
    )
    resource_server_url: Optional[AnyUrl] = Field(
        default=None,
        title="Resource Server URL",
        description="The URL for the resource server.",
    )



"""Pydantic Models for Resource Definitions, used to define default resources for a node or workcell."""

from typing import Annotated, Literal, Optional, Union

from madsci.common.types.base_types import BaseModel, new_ulid_str
from madsci.common.types.event_types import EventClientConfig
from madsci.common.types.lab_types import ManagerDefinition, ManagerType
from madsci.common.types.resource_types.custom_types import (
    CustomResourceTypes,
    ResourceTypeEnum,
)
from madsci.common.utils import new_name_str
from madsci.common.validators import ulid_validator
from pydantic.functional_validators import field_validator, model_validator
from pydantic.types import Tag
from sqlmodel import Field


class ResourceManagerDefinition(ManagerDefinition):
    """Definition for a Resource Manager's Configuration"""

    manager_type: Literal[ManagerType.RESOURCE_MANAGER] = Field(
        title="Manager Type",
        description="The type of the resource manager",
        default=ManagerType.RESOURCE_MANAGER,
    )
    host: str = Field(
        default="127.0.0.1",
        title="Server Host",
        description="The hostname or IP address of the Resource Manager server.",
    )
    port: int = Field(
        default=8003,
        title="Server Port",
        description="The port number of the Resource Manager server.",
    )
    db_url: str = Field(
        default="postgresql://rpl:rpl@localhost:5432/resources",
        title="Database URL",
        description="The URL of the database used by the Resource Manager.",
    )
    event_client_config: Optional[EventClientConfig] = Field(
        default=None,
        title="Event Client Configuration",
        description="Configuration for the event client.",
    )


class ResourceDefinition(BaseModel, table=False):
    """Definition for a MADSci Resource."""

    resource_id: str = Field(
        title="Resource ID",
        description="The ID of the resource.",
        default_factory=new_ulid_str,
        primary_key=True,
    )
    resource_name: str = Field(
        title="Resource Name",
        description="The name of the resource.",
        default_factory=new_name_str,
    )
    resource_type: str = Field(
        title="Resource Type",
        description="The type of the resource. Either a custom type name or a resource base type.",
        default="",
        nullable=False,
    )
    base_type: Literal[ResourceTypeEnum.resource] = Field(
        default=ResourceTypeEnum.resource,
        title="Resource Base Type",
        description="The base type of the resource.",
    )
    resource_description: Optional[str] = Field(
        default=None,
        title="Resource Description",
        description="A description of the resource.",
    )
    is_ulid = field_validator("resource_id")(ulid_validator)


class AssetResourceDefinition(ResourceDefinition, table=False):
    """Definition for an asset resource."""

    base_type: Literal[ResourceTypeEnum.asset] = Field(
        default=ResourceTypeEnum.asset,
        title="Resource Base Type",
        description="The base type of the asset.",
    )


class ConsumableResourceDefinition(ResourceDefinition):
    """Definition for a consumable resource."""

    base_type: Literal[ResourceTypeEnum.consumable] = Field(
        default=ResourceTypeEnum.consumable,
        title="Resource Base Type",
        description="The base type of the consumable.",
    )


class DiscreteConsumableResourceDefinition(ConsumableResourceDefinition):
    """Definition for a discrete consumable resource."""

    base_type: Literal[ResourceTypeEnum.discrete_consumable] = Field(
        default=ResourceTypeEnum.discrete_consumable,
        title="Resource Base Type",
        description="The base type of the consumable.",
    )


class ContinuousConsumableResourceDefinition(ConsumableResourceDefinition):
    """Definition for a continuous consumable resource."""

    base_type: Literal[ResourceTypeEnum.continuous_consumable] = Field(
        default=ResourceTypeEnum.continuous_consumable,
        title="Resource Base Type",
        description="The base type of the continuous consumable.",
    )


class ContainerResourceDefinition(ResourceDefinition):
    """Definition for a container resource."""

    base_type: Literal[ResourceTypeEnum.container] = Field(
        default=ResourceTypeEnum.container,
        title="Resource Base Type",
        description="The base type of the container.",
    )
    capacity: Optional[Union[int, float]] = Field(
        default=None,
        title="Container Capacity",
        description="The capacity of the container. If None, uses the type's default_capacity.",
    )
    default_children: Optional[
        Union[list[ResourceDefinition], dict[str, ResourceDefinition]]
    ] = Field(
        default=None,
        title="Default Children",
        description="The default children to create when initializing the container. If None, use the type's default_children.",
    )
    default_child_template: Optional[ResourceDefinition] = Field(
        default=None,
        title="Default Child Template",
        description="Template for creating child resources, supporting variable substitution. If None, use the type's default_child_template.",
    )


class CollectionResourceDefinition(ContainerResourceDefinition):
    """Definition for a collection resource. Collections are used for resources that have a number of children, each with a unique key, which can be randomly accessed."""

    base_type: Literal[ResourceTypeEnum.collection] = Field(
        default=ResourceTypeEnum.collection,
        title="Resource Base Type",
        description="The base type of the collection.",
    )
    keys: Optional[Union[int, list[str]]] = Field(
        default=None,
        title="Collection Keys",
        description="The keys for the collection. Can be an integer (converted to 1-based range) or explicit list.",
    )
    default_children: Optional[
        Union[list[ResourceDefinition], dict[str, ResourceDefinition]]
    ] = Field(
        default=None,
        title="Default Children",
        description="The default children to create when initializing the collection. If None, use the type's default_children.",
    )

    @field_validator("keys", mode="before")
    @classmethod
    def validate_keys(cls, v: Union[int, list[str], None]) -> Optional[list[str]]:
        """Convert integer keys to 1-based range if needed."""
        if isinstance(v, int):
            return [str(i) for i in range(1, v + 1)]
        return v


class RowResourceDefinition(ContainerResourceDefinition):
    """Definition for a row resource. Rows are 1D collections of resources. They are treated as single collections (i.e. Collection[Resource])."""

    base_type: Literal[ResourceTypeEnum.row] = Field(
        default=ResourceTypeEnum.row,
        title="Resource Base Type",
        description="The base type of the row.",
    )
    default_children: Optional[dict[str, ResourceDefinition]] = Field(
        default=None,
        title="Default Children",
        description="The default children to create when initializing the collection. If None, use the type's default_children.",
    )


class GridResourceDefinition(RowResourceDefinition):
    """Definition for a grid resource. Grids are 2D grids of resources. They are treated as nested collections (i.e. Collection[Collection[Resource]])."""

    base_type: Literal[ResourceTypeEnum.grid] = Field(
        default=ResourceTypeEnum.grid,
        title="Resource Base Type",
        description="The base type of the grid.",
    )
    default_children: Optional[dict[str, dict[str, ResourceDefinition]]] = Field(
        default=None,
        title="Default Children",
        description="The default children to create when initializing the collection. If None, use the type's default_children.",
    )
    column_dimension: int = Field(
        default=None,
        title="Column Dimension",
        description="The number of columns in the grid. If None, use the type's column_dimension.",
    )


class VoxelGridResourceDefinition(GridResourceDefinition):
    """Definition for a voxel grid resource. Voxel grids are 3D grids of resources. They are treated as nested collections (i.e. Collection[Collection[Collection[Resource]]])."""

    base_type: Literal[ResourceTypeEnum.voxel_grid] = Field(
        default=ResourceTypeEnum.voxel_grid,
        title="Resource Base Type",
        description="The base type of the voxel grid.",
    )
    default_children: Optional[dict[str, dict[str, dict[str, ResourceDefinition]]]] = (
        Field(
            default=None,
            title="Default Children",
            description="The default children to create when initializing the collection. If None, use the type's default_children.",
        )
    )
    layer_dimension: int = Field(
        title="Layer Dimension",
        description="The number of layers in the voxel grid. If None, use the type's layer_dimension.",
    )


class SlotResourceDefinition(ContainerResourceDefinition):
    """Definition for a slot resource."""

    base_type: Literal[ResourceTypeEnum.slot] = Field(
        default=ResourceTypeEnum.slot,
        title="Resource Base Type",
        description="The base type of the slot.",
    )
    default_child_quantity: Optional[int] = Field(
        default=None,
        title="Default Child Quantity",
        description="The number of children to create by default. If None, use the type's default_child_quantity.",
        ge=0,
        le=1,
    )


class StackResourceDefinition(ContainerResourceDefinition):
    """Definition for a stack resource."""

    base_type: Literal[ResourceTypeEnum.stack] = Field(
        default=ResourceTypeEnum.stack,
        title="Resource Base Type",
        description="The base type of the stack.",
    )
    default_child_quantity: Optional[int] = Field(
        default=None,
        title="Default Child Quantity",
        description="The number of children to create by default. If None, use the type's default_child_quantity.",
    )


class QueueResourceDefinition(ContainerResourceDefinition):
    """Definition for a queue resource."""

    base_type: Literal[ResourceTypeEnum.queue] = Field(
        default=ResourceTypeEnum.queue,
        title="Resource Base Type",
        description="The base type of the queue.",
    )
    default_child_quantity: Optional[int] = Field(
        default=None,
        title="Default Child Quantity",
        description="The number of children to create by default. If None, use the type's default_child_quantity.",
    )


class PoolResourceDefinition(ContainerResourceDefinition):
    """Definition for a pool resource. Pool resources are collections of consumables with no structure (used for wells, reservoirs, etc.)."""

    base_type: Literal[ResourceTypeEnum.pool] = Field(
        default=ResourceTypeEnum.pool,
        title="Resource Base Type",
        description="The base type of the pool.",
    )


ResourceDefinitions = Union[
    Annotated[ResourceDefinition, Tag("resource")],
    Annotated[AssetResourceDefinition, Tag("asset")],
    Annotated[ContainerResourceDefinition, Tag("container")],
    Annotated[CollectionResourceDefinition, Tag("collection")],
    Annotated[GridResourceDefinition, Tag("grid")],
    Annotated[VoxelGridResourceDefinition, Tag("voxel_grid")],
    Annotated[StackResourceDefinition, Tag("stack")],
    Annotated[QueueResourceDefinition, Tag("queue")],
    Annotated[PoolResourceDefinition, Tag("pool")],
]


class ResourceFile(BaseModel):
    """Definition for a MADSci Resource File."""

    resource_types: list[
        Annotated[CustomResourceTypes, Field(discriminator="base_type")]
    ] = Field(
        title="Resource Types",
        description="The definitions of the resource types in the file.",
        default=[],
    )
    default_resources: list[
        Annotated[ResourceDefinitions, Field(discriminator="base_type")]
    ] = Field(
        title="Default Resources",
        description="The definitions of the default resources in the file.",
        default=[],
    )

    @model_validator(mode="after")
    def validate_resource_types(self) -> "ResourceFile":
        """Validate resource types."""
        for resource_type in self.resource_types:
            for parent_type in resource_type.parent_types:
                if parent_type not in ResourceTypeEnum and parent_type not in [
                    resource_type.type_name for resource_type in self.resource_types
                ]:
                    raise ValueError(
                        f"Unknown resource parent type: {parent_type}, parent type must be one of {list(ResourceTypeEnum)} or a custom defined resource type.",
                    )
        return self



"""Types related to MADSci Resources."""

import string
from typing import Annotated, Literal, Optional, Union

from madsci.common.types.resource_types.custom_types import (
    AssetTypeEnum,
    ConsumableTypeEnum,
    ContainerTypeEnum,
    ResourceTypeEnum,
)
from madsci.common.types.resource_types.definitions import (
    AssetResourceDefinition,
    CollectionResourceDefinition,
    ConsumableResourceDefinition,
    ContainerResourceDefinition,
    ContinuousConsumableResourceDefinition,
    DiscreteConsumableResourceDefinition,
    GridResourceDefinition,
    PoolResourceDefinition,
    QueueResourceDefinition,
    ResourceDefinition,
    RowResourceDefinition,
    SlotResourceDefinition,
    StackResourceDefinition,
    VoxelGridResourceDefinition,
)
from pydantic import (
    AfterValidator,
    AnyUrl,
    computed_field,
    model_validator,
)
from pydantic.types import Discriminator, Tag, datetime
from sqlalchemy.dialects.postgresql import JSON
from sqlalchemy.sql.sqltypes import String
from sqlmodel import Field
from typing_extensions import Self  # type: ignore

PositiveInt = Annotated[int, Field(ge=0)]
PositiveNumber = Annotated[Union[float, int], Field(ge=0)]


class Resource(ResourceDefinition, extra="allow", table=False):
    """Base class for all MADSci Resources. Used to track any resource that isn't well-modeled by a more specific type."""

    resource_url: Optional[AnyUrl] = Field(
        title="Resource URL",
        description="The URL of the resource.",
        nullable=True,
        default=None,
        sa_type=String,
    )
    base_type: Literal[ResourceTypeEnum.resource] = Field(
        title="Resource Base Type",
        description="The base type of the resource.",
        nullable=False,
        default=ResourceTypeEnum.resource,
        sa_type=String,
    )
    parent_id: Optional[str] = Field(
        default=None,
        title="Parent Resource",
        description="The parent resource ID, if any.",
    )
    key: Optional[str] = Field(
        default=None,
        title="Key",
        description="The key of the resource in the parent container, if any.",
    )
    owner: Optional[str] = Field(
        title="Resource owner",
        description="The owner of the resource.",
        nullable=True,
        default=None,
    )
    attributes: dict = Field(
        default_factory=dict,
        sa_type=JSON,
        title="Attributes",
        description="Custom attributes for the asset.",
    )
    created_at: Optional[datetime] = Field(
        title="Created Datetime",
        description="The timestamp of when the resource was created.",
        default=None,
    )
    updated_at: Optional[datetime] = Field(
        title="Updated Datetime",
        description="The timestamp of when the resource was last updated.",
        default=None,
    )
    removed: bool = Field(
        title="Removed",
        description="Whether the resource has been removed from the lab.",
        nullable=False,
        default=False,
    )

    @classmethod
    def discriminate(cls, resource: "ResourceDataModels") -> "ResourceDataModels":
        """Discriminate the resource based on its base type."""
        if isinstance(resource, dict):
            resource_type = resource.get("base_type")
        else:
            resource_type = resource.base_type
        return RESOURCE_TYPE_MAP[resource_type]["model"].model_validate(resource)


class Asset(Resource):
    """Base class for all MADSci Assets. These are tracked resources that aren't consumed (things like samples, labware, etc.)."""

    base_type: Literal[AssetTypeEnum.asset] = Field(
        title="Asset Base Type",
        description="The base type of the asset.",
        nullable=False,
        default=AssetTypeEnum.asset,
    )


class Consumable(Resource):
    """Base class for all MADSci Consumables. These are resources that are consumed (things like reagents, pipette tips, etc.)."""

    base_type: Literal[ConsumableTypeEnum.consumable] = Field(
        title="Consumable Base Type",
        description="The base type of the consumable.",
        nullable=False,
        default=ConsumableTypeEnum.consumable,
    )
    quantity: PositiveNumber = Field(
        title="Quantity",
        description="The quantity of the consumable.",
    )
    capacity: Optional[PositiveNumber] = Field(
        title="Capacity",
        description="The maximum capacity of the consumable.",
        default=None,
    )

    @model_validator(mode="after")
    def validate_consumable_quantity(self) -> Self:
        """Validate that the quantity is less than or equal to the capacity."""
        if self.capacity is not None and self.quantity > self.capacity:
            raise ValueError("Quantity cannot be greater than capacity.")
        return self


class DiscreteConsumable(Consumable):
    """Base class for all MADSci Discrete Consumables. These are consumables that are counted in whole numbers (things like pipette tips, tubes, etc.)."""

    base_type: Literal[ConsumableTypeEnum.discrete_consumable] = Field(
        title="Consumable Base Type",
        description="The base type of the discrete consumable.",
        default=ConsumableTypeEnum.discrete_consumable,
        const=True,
    )
    quantity: PositiveInt = Field(
        title="Quantity",
        description="The quantity of the discrete consumable.",
    )
    capacity: Optional[PositiveInt] = Field(
        title="Capacity",
        description="The maximum capacity of the discrete consumable.",
        default=None,
    )


class ContinuousConsumable(Consumable):
    """Base class for all MADSci Continuous Consumables. These are consumables that are measured in continuous quantities (things like liquids, powders, etc.)."""

    base_type: Literal[ConsumableTypeEnum.continuous_consumable] = Field(
        title="Consumable Base Type",
        description="The base type of the continuous consumable.",
        default=ConsumableTypeEnum.continuous_consumable,
        const=True,
    )
    quantity: PositiveNumber = Field(
        title="Quantity",
        description="The quantity of the continuous consumable.",
    )
    capacity: Optional[PositiveNumber] = Field(
        title="Capacity",
        description="The maximum capacity of the continuous consumable.",
        default=None,
    )


class Container(Asset):
    """Data Model for a Container. A container is a resource that can hold other resources."""

    base_type: Literal[ContainerTypeEnum.container] = Field(
        title="Container Base Type",
        description="The base type of the container.",
        nullable=False,
        default=ContainerTypeEnum.container,
    )
    capacity: Optional[PositiveInt] = Field(
        title="Capacity",
        description="The capacity of the container.",
        default=None,
    )
    children: Optional[dict[str, "ResourceDataModels"]] = Field(
        title="Children",
        description="The children of the container.",
        default_factory=dict,
    )

    def get_child(self, key: str) -> Optional["ResourceDataModels"]:
        """Get a child from the container."""
        return self.children.get(key, None)

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return len(self.children)

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the container as a flat dictionary."""
        return self.children

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the container."""
        self.children = children

    @model_validator(mode="after")
    def validate_container_quantity(self) -> Self:
        """Validate that the quantity is less than or equal to the capacity."""
        if self.capacity is not None and self.quantity > self.capacity:
            raise ValueError("Quantity cannot be greater than capacity.")
        return self


class Collection(Container):
    """Data Model for a Collection. A collection is a container that can hold other resources, and which supports random access."""

    base_type: Literal[ContainerTypeEnum.collection] = Field(
        title="Container Base Type",
        description="The base type of the collection.",
        default=ContainerTypeEnum.collection,
        const=True,
    )
    children: Optional[dict[str, "ResourceDataModels"]] = Field(
        title="Children",
        description="The children of the collection.",
        default_factory=dict,
    )

    def get_child(self, key: str) -> Optional["ResourceDataModels"]:
        """Get a child from the container."""
        return self.children.get(key, None)

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return len(self.children)

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the collection as a flat dictionary."""
        return self.children

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the collection."""
        self.children = children


def single_letter_or_digit_validator(value: str) -> str:
    """Validate that the value is a single letter or digit."""
    if not (value.isalpha() and len(value) == 1) or value.isdigit():
        raise ValueError("Value must be a single letter or digit.")
    return value


GridIndex = Union[
    int,
    Annotated[str, AfterValidator(single_letter_or_digit_validator)],
]
GridIndex2D = tuple[GridIndex, GridIndex]
GridIndex3D = tuple[GridIndex, GridIndex, GridIndex]


def numericize_index(key: Union[str, GridIndex]) -> GridIndex:
    """Convert a key to a numeric value."""
    if isinstance(key, int) or str(key).isdigit():
        return int(key)
    return string.ascii_lowercase.index(key.lower())


class Row(Container):
    """Data Model for a Row. A row is a container that can hold other resources in a single dimension and supports random access. For example, a row of tubes in a rack or a single-row microplate. Rows are indexed by integers or letters."""

    children: Optional[dict[GridIndex, "ResourceDataModels"]] = Field(
        title="Children",
        description="The children of the row container.",
        default_factory=dict,
    )
    base_type: Literal[ContainerTypeEnum.row] = Field(
        title="Container Base Type",
        description="The base type of the row.",
        default=ContainerTypeEnum.row,
        const=True,
    )
    row_dimension: int = Field(
        title="Row Dimension",
        description="The number of rows in the row.",
        ge=0,
    )

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return len(self.children)

    def get_child(self, key: GridIndex) -> Optional["ResourceDataModels"]:
        """Get a child from the Row."""
        return self.children.get(numericize_index(key), None)

    @staticmethod
    def flatten_key(key: GridIndex) -> str:
        """Flatten the key to a string."""
        return str(key)

    @staticmethod
    def expand_key(key: str) -> GridIndex:
        """Expand the key to a 2-tuple."""
        if key.isdigit():
            return int(key)
        return key

    def check_key_bounds(self, key: Union[str, GridIndex]) -> bool:
        """Check if the key is within the bounds of the grid."""
        key = numericize_index(key)
        return not (key < 0 or key >= self.row_dimension)


class Grid(Row):
    """Data Model for a Grid. A grid is a container that can hold other resources in two dimensions and supports random access. For example, a 96-well microplate. Grids are indexed by integers or letters."""

    children: Optional[dict[GridIndex, dict[GridIndex, "ResourceDataModels"]]] = Field(
        title="Children",
        description="The children of the grid container.",
        default_factory=dict,
    )
    base_type: Literal[ContainerTypeEnum.grid] = Field(
        title="Container Base Type",
        description="The base type of the grid.",
        default=ContainerTypeEnum.grid,
        const=True,
    )
    column_dimension: int = Field(
        title="Column Dimension",
        description="The number of columns in the grid.",
        ge=0,
    )

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        quantity = 0
        for _, row_value in self.children.items():
            quantity += len(row_value)
        return quantity

    def get_child(self, key: GridIndex2D) -> Optional["ResourceDataModels"]:
        """Get a child from the Grid."""
        row = self.children.get(numericize_index(key[0]), None)
        if row is None:
            return None
        return row.get(numericize_index(key[1]), None)

    @staticmethod
    def flatten_key(key: GridIndex2D) -> str:
        """Flatten the key to a string."""
        return "_".join([str(index) for index in key])

    @staticmethod
    def expand_key(key: str) -> GridIndex2D:
        """Expand the key to a 2-tuple."""
        expanded_key = key.split("_")
        final_keys = []
        for index in expanded_key:
            if index.isdigit():
                final_keys.append(int(index))
            else:
                final_keys.append(index)
        return tuple(final_keys)

    def check_key_bounds(self, key: Union[str, GridIndex2D]) -> bool:
        """Check if the key is within the bounds of the grid."""
        if isinstance(key, str):
            key = self.expand_key(key)
        elif not isinstance(key, tuple) or len(key) != 2:
            raise ValueError("Key must be a string or a 2-tuple.")
        numeric_key = []
        for index in key:
            numeric_key.append(numericize_index(index))
        key = numeric_key
        return not (key[0] < 0 or key[0] >= self.row_dimension) and not (
            key[1] < 0 or key[1] >= self.column_dimension
        )

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the grid as a flat dictionary."""
        children_dict = {}
        for row_key, row_value in self.children.items():
            for col_key, col_value in row_value.items():
                children_dict[self.flatten_key((row_key, col_key))] = col_value

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the grid."""
        for key, value in children.items():
            row_key, col_key = self.expand_key(key)
            if row_key not in self.children:
                self.children[row_key] = {}
            self.children[row_key][col_key] = value


class VoxelGrid(Grid):
    """Data Model for a Voxel Grid. A voxel grid is a container that can hold other resources in three dimensions and supports random access. Voxel grids are indexed by integers or letters."""

    base_type: Literal[ContainerTypeEnum.voxel_grid] = Field(
        title="Container Base Type",
        description="The base type of the voxel grid.",
        default=ContainerTypeEnum.voxel_grid,
        const=True,
    )
    layer_dimension: int = Field(
        title="Layer Dimension",
        description="The number of layers in the grid.",
        ge=0,
    )
    children: Optional[
        dict[GridIndex, dict[GridIndex, dict[GridIndex, "ResourceDataModels"]]]
    ] = Field(
        title="Children",
        description="The children of the voxel grid container.",
        default_factory=dict,
    )

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        quantity = 0
        for _, row_value in self.children.items():
            for _, col_value in row_value.items():
                quantity += len(col_value)
        return quantity

    def get_child(self, key: GridIndex3D) -> Optional["ResourceDataModels"]:
        """Get a child from the Voxel Grid."""
        row = self.children.get(numericize_index(key[0]), None)
        if row is None:
            return None
        col = row.get(numericize_index(key[1]), None)
        if col is None:
            return None
        return col.get(numericize_index(key[2]), None)

    @staticmethod
    def flatten_key(key: GridIndex3D) -> str:
        """Flatten the key to a string."""
        return "_".join([str(index) for index in key])

    @staticmethod
    def expand_key(key: str) -> GridIndex3D:
        """Expand the key to a tuple."""
        expanded_key = key.split("_")
        final_keys = []
        for index in expanded_key:
            if index.isdigit():
                final_keys.append(int(index))
            else:
                final_keys.append(index)
        return final_keys

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the grid as a flat dictionary."""
        children_dict = {}
        for row_key, row_value in self.children.items():
            for col_key, col_value in row_value.items():
                for depth_key, depth_value in col_value.items():
                    children_dict[self.flatten_key((row_key, col_key, depth_key))] = (
                        depth_value
                    )

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the grid."""
        for key, value in children.items():
            row_key, col_key, depth_key = self.expand_key(key)
            if row_key not in self.children:
                self.children[row_key] = {}
            if col_key not in self.children[row_key]:
                self.children[row_key][col_key] = {}
            self.children[row_key][col_key][depth_key] = value

    def check_key_bounds(self, key: Union[str, GridIndex3D]) -> bool:
        """Check if the key is within the bounds of the grid."""
        if isinstance(key, str):
            key = self.expand_key(key)
        elif not isinstance(key, tuple) or len(key) != 3:
            raise ValueError("Key must be a string or a 3-tuple.")
        numeric_key = []
        for index in key:
            numeric_key.append(numericize_index(index))
        key = numeric_key
        return (
            not (key[0] < 0 or key[0] >= self.row_dimension)
            and not (key[1] < 0 or key[1] >= self.column_dimension)
            and not (key[2] < 0 or key[2] >= self.layer_dimension)
        )


class Slot(Container):
    """Data Model for a Slot. A slot is a container that can hold a single resource."""

    base_type: Literal[ContainerTypeEnum.slot] = Field(
        title="Container Base Type",
        description="The base type of the slot.",
        default=ContainerTypeEnum.slot,
        const=True,
    )
    children: Optional[list["ResourceDataModels"]] = Field(
        title="Children",
        description="The children of the slot.",
        default_factory=list,
    )
    capacity: Literal[1] = Field(
        title="Capacity",
        description="The capacity of the slot.",
        default=1,
        const=1,
    )

    def get_child(self, key: Optional[int] = None) -> Optional["ResourceDataModels"]:
        """Get the child from the slot."""
        if key is not None and key != 0:
            return None
        return self.child

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return len(self.children)

    @property
    def child(self) -> Optional["ResourceDataModels"]:
        """Get the child from the slot."""
        return self.children[0] if self.children else None

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the stack as a flat dict."""
        return {"0": self.child} if self.child else {}

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the stack."""
        self.children = [children["0"]] if "0" in children else []


class Stack(Container):
    """Data Model for a Stack. A stack is a container that can hold other resources in a single dimension and supports last-in, first-out (LIFO) access. For example, a stack of plates in a vertical magazine. Stacks are indexed by integers, with 0 being the bottom."""

    base_type: Literal[ContainerTypeEnum.stack] = Field(
        title="Container Base Type",
        description="The base type of the stack.",
        default=ContainerTypeEnum.stack,
        const=True,
    )
    children: Optional[list["ResourceDataModels"]] = Field(
        title="Children",
        description="The children of the stack.",
        default_factory=list,
    )

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return len(self.children)

    def get_child(self, key: int) -> Optional["ResourceDataModels"]:
        """Get a child from the container."""
        key = int(key)
        if key < 0 or key >= len(self.children):
            return None
        return self.children[key]

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the stack as a flat dict."""
        children_dict = {}
        for i in range(len(self.children)):
            children_dict[str(i)] = self.children[i]
        return children_dict

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the stack."""
        ordered_children = sorted(children.items(), key=lambda x: int(x[0]))
        self.children = [child[1] for child in ordered_children]


class Queue(Container):
    """Data Model for a Queue. A queue is a container that can hold other resources in a single dimension and supports first-in, first-out (FIFO) access. For example, a conveyer belt. Queues are indexed by integers, with 0 being the front."""

    base_type: Literal[ContainerTypeEnum.queue] = Field(
        title="Container Base Type",
        description="The base type of the queue.",
        default=ContainerTypeEnum.queue,
        const=True,
    )
    children: Optional[list["ResourceDataModels"]] = Field(
        title="Children",
        description="The children of the queue.",
        default_factory=list,
    )

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return len(self.children)

    def get_child(self, key: int) -> Optional["ResourceDataModels"]:
        """Get a child from the container."""
        key = int(key)
        if key < 0 or key >= len(self.children):
            return None
        return self.children[key]

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the stack as a flat dict."""
        children_dict = {}
        for i in range(len(self.children)):
            children_dict[str(i)] = self.children[i]
        return children_dict

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the queue."""
        ordered_children = sorted(children.items(), key=lambda x: int(x[0]))
        self.children = [child[1] for child in ordered_children]


class Pool(Container):
    """Data Model for a Pool. A pool is a container for holding consumables that can be mixed or collocated. For example, a single well in a microplate, or a reservoir. Pools are indexed by string key."""

    base_type: Literal[ContainerTypeEnum.pool] = Field(
        title="Container Base Type",
        description="The base type of the pool.",
        default=ContainerTypeEnum.pool,
        const=True,
    )
    children: Optional[dict[str, "ConsumableDataModels"]] = Field(
        title="Children",
        description="The children of the pool.",
        default_factory=dict,
    )
    capacity: Optional[float] = Field(
        title="Capacity",
        description="The capacity of the pool as a whole.",
    )

    def get_child(self, key: str) -> Optional["ResourceDataModels"]:
        """Get a child from the container."""
        return self.children.get(key, None)

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return sum(
            [
                child.quantity
                for child in self.children.values()
                if hasattr(child, "quantity")
            ]
        )

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the pool as a flat dictionary."""
        return self.children

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the pool."""
        self.children = children


RESOURCE_TYPE_MAP = {
    ResourceTypeEnum.resource: {
        "definition": ResourceDefinition,
        "model": Resource,
    },
    ResourceTypeEnum.asset: {
        "definition": AssetResourceDefinition,
        "model": Asset,
    },
    ResourceTypeEnum.consumable: {
        "definition": ConsumableResourceDefinition,
        "model": Consumable,
    },
    ConsumableTypeEnum.discrete_consumable: {
        "definition": DiscreteConsumableResourceDefinition,
        "model": DiscreteConsumable,
    },
    ConsumableTypeEnum.continuous_consumable: {
        "definition": ContinuousConsumableResourceDefinition,
        "model": ContinuousConsumable,
    },
    AssetTypeEnum.container: {
        "definition": ContainerResourceDefinition,
        "model": Container,
    },
    ContainerTypeEnum.stack: {
        "definition": StackResourceDefinition,
        "model": Stack,
    },
    ContainerTypeEnum.queue: {
        "definition": QueueResourceDefinition,
        "model": Queue,
    },
    ContainerTypeEnum.collection: {
        "definition": CollectionResourceDefinition,
        "model": Collection,
    },
    ContainerTypeEnum.row: {
        "definition": RowResourceDefinition,
        "model": Row,
    },
    ContainerTypeEnum.grid: {
        "definition": GridResourceDefinition,
        "model": Grid,
    },
    ContainerTypeEnum.voxel_grid: {
        "definition": VoxelGridResourceDefinition,
        "model": VoxelGrid,
    },
    ContainerTypeEnum.pool: {
        "definition": PoolResourceDefinition,
        "model": Pool,
    },
    ContainerTypeEnum.slot: {
        "definition": SlotResourceDefinition,
        "model": Slot,
    },
}

ResourceDataModels = Annotated[
    Union[
        Annotated[Resource, Tag("resource")],
        Annotated[Asset, Tag("asset")],
        Annotated[Consumable, Tag("consumable")],
        Annotated[DiscreteConsumable, Tag("discrete_consumable")],
        Annotated[ContinuousConsumable, Tag("continuous_consumable")],
        Annotated[Container, Tag("container")],
        Annotated[Collection, Tag("collection")],
        Annotated[Row, Tag("row")],
        Annotated[Grid, Tag("grid")],
        Annotated[VoxelGrid, Tag("voxel_grid")],
        Annotated[Stack, Tag("stack")],
        Annotated[Queue, Tag("queue")],
        Annotated[Pool, Tag("pool")],
        Annotated[Slot, Tag("slot")],
    ],
    Discriminator("base_type"),
]

ConsumableDataModels = Annotated[
    Union[
        Annotated[Consumable, Tag("consumable")],
        Annotated[DiscreteConsumable, Tag("discrete_consumable")],
        Annotated[ContinuousConsumable, Tag("continuous_consumable")],
    ],
    Discriminator("base_type"),
]

ContainerDataModels = Annotated[
    Union[
        Annotated[Container, Tag("container")],
        Annotated[Collection, Tag("collection")],
        Annotated[Row, Tag("row")],
        Annotated[Grid, Tag("grid")],
        Annotated[VoxelGrid, Tag("voxel_grid")],
        Annotated[Stack, Tag("stack")],
        Annotated[Queue, Tag("queue")],
        Annotated[Pool, Tag("pool")],
        Annotated[Slot, Tag("slot")],
    ],
    Discriminator("base_type"),
]



"""Data Models for validating Custom Resource Type Definitions"""

from enum import Enum
from typing import TYPE_CHECKING, Annotated, Any, Literal, Optional, Union

from madsci.common.types.base_types import BaseModel

if TYPE_CHECKING:
    from madsci.common.types.resource_types.definitions import ResourceDefinition
from pydantic import Json, Tag
from pydantic.config import ConfigDict
from pydantic.functional_validators import field_validator
from sqlmodel import Field


class AssetTypeEnum(str, Enum):
    """Type for a MADSci Asset."""

    container = "container"
    asset = "asset"


class ConsumableTypeEnum(str, Enum):
    """Type for a MADSci Consumable."""

    consumable = "consumable"
    discrete_consumable = "discrete_consumable"
    continuous_consumable = "continuous_consumable"


class ContainerTypeEnum(str, Enum):
    """Type for a MADSci Container."""

    container = "container"
    slot = "slot"
    stack = "stack"
    queue = "queue"
    collection = "collection"
    row = "row"
    grid = "grid"
    voxel_grid = "voxel_grid"
    pool = "pool"


class ResourceTypeEnum(str, Enum):
    """Enum for all resource base types."""

    """Resource Base Types"""
    resource = "resource"

    """Asset Resource Base Types"""
    asset = "asset"
    container = "container"

    """Consumable Resource Base Types"""
    consumable = "consumable"
    discrete_consumable = "discrete_consumable"
    continuous_consumable = "continuous_consumable"

    """Container Resource Base Types"""
    slot = "slot"
    stack = "stack"
    queue = "queue"
    collection = "collection"
    row = "row"
    grid = "grid"
    voxel_grid = "voxel_grid"
    pool = "pool"


class CustomResourceAttributeDefinition(BaseModel, extra="allow"):
    """Definition for a MADSci Custom Resource Attribute."""

    attribute_name: str = Field(
        title="Attribute Name",
        description="The name of the attribute.",
    )
    attribute_description: Optional[str] = Field(
        default=None,
        title="Attribute Description",
        description="A description of the attribute.",
    )
    optional: bool = Field(
        default=False,
        title="Optional",
        description="Whether the attribute is optional.",
    )
    default_value: Json[Any] = Field(
        default=None,
        title="Default Value",
        description="The default value of the attribute.",
    )


class ResourceTypeDefinition(BaseModel):
    """Definition for a MADSci Resource Type."""

    model_config = ConfigDict(extra="allow")

    type_name: str = Field(
        title="Resource Type Name",
        description="The name of the type of resource (i.e. 'plate_96_well_corningware', 'tube_rack_24', etc.).",
    )
    type_description: str = Field(
        title="Resource Type Description",
        description="A description of the custom type of the resource.",
    )
    base_type: Literal[ResourceTypeEnum.resource] = Field(
        default=ResourceTypeEnum.resource,
        title="Resource Base Type",
        description="The base type of the resource.",
    )
    parent_types: list[str] = Field(
        default=["resource"],
        title="Resource Parent Types",
        description="The parent types of the resource.",
    )
    custom_attributes: Optional[list["CustomResourceAttributeDefinition"]] = Field(
        default=None,
        title="Custom Attributes",
        description="Custom attributes used by resources of this type.",
    )

    @field_validator("parent_types", mode="before")
    @classmethod
    def validate_parent_types(cls, v: Union[list[str], str]) -> list[str]:
        """Validate parent types."""
        if isinstance(v, str):
            return [v]
        return v


class ContainerResourceTypeDefinition(ResourceTypeDefinition):
    """Definition for a MADSci Container Resource Type."""

    supported_child_types: list[str] = Field(
        title="Supported Child Types",
        description="The resource types for children supported by the container. If `resource` is included, the container can contain any resource type.",
    )
    default_capacity: Optional[Union[int, float]] = Field(
        title="Default Capacity",
        description="The default maximum capacity of the container. If None, the container has no capacity limit.",
        default=None,
    )
    resizeable: bool = Field(
        default=False,
        title="Resizeable",
        description="Whether containers of this type support different sizes. If True, the container can be resized. If False, the container is fixed size.",
    )
    default_children: Optional[
        Union[list["ResourceDefinition"], dict[str, "ResourceDefinition"]]
    ] = Field(
        default=None,
        title="Default Children",
        description="The default children to create when populating the container. Takes precedence over default_child_template.",
    )
    default_child_template: Optional[list["ResourceDefinition"]] = Field(
        default=None,
        title="Default Child Template",
        description="The default template for children to create when populating the container.",
    )
    base_type: Literal[ResourceTypeEnum.container] = Field(
        default=ResourceTypeEnum.container,
        title="Container Base Type",
        description="The base type of the container.",
    )


class AssetResourceTypeDefinition(ResourceTypeDefinition):
    """Definition for a MADSci Asset Resource Type."""

    base_type: Literal[ResourceTypeEnum.asset] = Field(
        default=ResourceTypeEnum.asset,
        title="Asset Base Type",
        description="The base type of the asset.",
    )


class ConsumableResourceTypeDefinition(ResourceTypeDefinition):
    """Definition for a MADSci Consumable Resource Type."""

    base_type: Literal[ResourceTypeEnum.consumable] = Field(
        default=ResourceTypeEnum.consumable,
        title="Consumable Base Type",
        description="The base type of the consumable.",
    )


class DiscreteConsumableResourceTypeDefinition(ConsumableResourceTypeDefinition):
    """Definition for a MADSci Discrete Consumable Resource Type."""

    base_type: Literal[ResourceTypeEnum.discrete_consumable] = Field(
        default=ResourceTypeEnum.discrete_consumable,
        title="Discrete Consumable Base Type",
        description="The base type of the discrete consumable.",
    )


class ContinuousConsumableResourceTypeDefinition(ConsumableResourceTypeDefinition):
    """Definition for a MADSci Continuous Consumable Resource Type."""

    base_type: Literal[ResourceTypeEnum.continuous_consumable] = Field(
        default=ResourceTypeEnum.continuous_consumable,
        title="Continuous Consumable Base Type",
        description="The base type of the continuous consumable.",
    )


class StackResourceTypeDefinition(ContainerResourceTypeDefinition):
    """Definition for a MADSci Stack Resource Type."""

    default_child_quantity: Optional[int] = Field(
        default=None,
        title="Default Child Quantity",
        description="The default number of children to create when populating the container. If None, the container will be populated with a single child.",
    )
    base_type: Literal[ResourceTypeEnum.stack] = Field(
        default=ResourceTypeEnum.stack,
        title="Stack Base Type",
        description="The base type of the stack.",
    )


class QueueResourceTypeDefinition(ContainerResourceTypeDefinition):
    """Definition for a MADSci Queue Resource Type."""

    default_child_quantity: Optional[int] = Field(
        default=None,
        title="Default Child Quantity",
        description="The default number of children to create when populating the container. If None, the container will be populated with a single child.",
    )
    base_type: Literal[ResourceTypeEnum.queue] = Field(
        default=ResourceTypeEnum.queue,
        title="Queue Base Type",
        description="The base type of the queue.",
    )


class CollectionResourceTypeDefinition(ContainerResourceTypeDefinition):
    """Definition for a MADSci Collection Resource Type."""

    keys: Optional[list[str]] = Field(
        title="Collection Keys",
        description="The keys of the collection.",
    )
    default_children: Optional[
        Union[list["ResourceDefinition"], dict[str, "ResourceDefinition"]]
    ] = Field(
        default=None,
        title="Default Children",
        description="The default children to create when populating the container.",
    )
    base_type: Literal[ResourceTypeEnum.collection] = Field(
        default=ResourceTypeEnum.collection,
        title="Collection Base Type",
        description="The base type of the collection.",
    )

    @field_validator("keys", mode="before")
    @classmethod
    def validate_keys(cls, v: Union[int, list[str]]) -> list[str]:
        """Convert integer keys count to 1-indexed range."""
        if isinstance(v, int):
            return [str(i) for i in range(1, v + 1)]
        return v


class RowResourceTypeDefinition(ContainerResourceTypeDefinition):
    """Definition for a MADSci Row Resource Type."""

    base_type: Literal[ResourceTypeEnum.row] = Field(
        default=ResourceTypeEnum.row,
        title="Row Base Type",
        description="The base type of the row.",
    )


class GridResourceTypeDefinition(ContainerResourceTypeDefinition):
    """Definition for a MADSci Grid Resource Type."""

    rows: list[str] = Field(
        title="Grid Rows",
        description="The row labels for the grid.",
    )
    columns: list[str] = Field(
        title="Grid Columns",
        description="The column labels for the grid.",
    )
    base_type: Literal[ResourceTypeEnum.grid] = Field(
        default=ResourceTypeEnum.grid,
        title="Grid Base Type",
        description="The base type of the grid.",
    )

    @field_validator("columns", "rows", mode="before")
    @classmethod
    def validate_keys(cls, v: Union[int, list[str]]) -> list[str]:
        """Convert integer keys count to 1-indexed range."""
        if isinstance(v, int):
            return [str(i) for i in range(1, v + 1)]
        return v


class VoxelGridResourceTypeDefinition(GridResourceTypeDefinition):
    """Definition for a MADSci Voxel Grid Resource Type."""

    capacity: Optional[int] = Field(
        title="Collection Capacity",
        description="The maximum capacity of each element in the grid.",
    )
    planes: list[str] = Field(
        title="Voxel Grid Planes",
        description="The keys of the planes in the grid.",
    )
    base_type: Literal[ResourceTypeEnum.voxel_grid] = Field(
        default=ResourceTypeEnum.voxel_grid,
        title="Voxel Grid Base Type",
        description="The base type of the voxel grid.",
    )

    @field_validator("columns", "rows", mode="before")
    @classmethod
    def validate_keys(cls, v: Union[int, list[str]]) -> list[str]:
        """Convert integer keys count to 1-indexed range."""
        if isinstance(v, int):
            return [str(i) for i in range(1, v + 1)]
        return v


class PoolResourceTypeDefinition(ContainerResourceTypeDefinition):
    """Definition for a MADSci Pool Resource Type."""

    base_type: Literal[ResourceTypeEnum.pool] = Field(
        default=ResourceTypeEnum.pool,
        title="Pool Base Type",
        description="The base type of the pool.",
    )


class SlotTypeDefinition(ContainerResourceTypeDefinition):
    """Definition for a MADSci Slot Resource Type."""

    base_type: Literal[ResourceTypeEnum.slot] = Field(
        default=ResourceTypeEnum.slot,
        title="Slot Base Type",
        description="The base type of the slot.",
    )


CustomResourceTypes = Union[
    Annotated[ResourceTypeDefinition, Tag("resource")],  # * resource: Resource
    Annotated[
        ContainerResourceTypeDefinition, Tag("container")
    ],  # * container of resources: Container[Resource]
    Annotated[AssetResourceTypeDefinition, Tag("asset")],  # * trackable resource: Asset
    Annotated[
        ConsumableResourceTypeDefinition, Tag("consumable")
    ],  # * consumable resource: Consumable
    Annotated[
        StackResourceTypeDefinition, Tag("stack")
    ],  # * stack of resources: Container[Resource]
    Annotated[
        QueueResourceTypeDefinition, Tag("queue")
    ],  # * queue of resources: Container[Resource]
    Annotated[
        CollectionResourceTypeDefinition, Tag("collection")
    ],  # * collection of resources: Container[Resource]
    Annotated[
        RowResourceTypeDefinition, Tag("row")
    ],  # * row of resources: Collection[Resource]
    Annotated[
        GridResourceTypeDefinition, Tag("grid")
    ],  # * 2D grid of resources: Collection[Collection[Resource]]
    Annotated[
        VoxelGridResourceTypeDefinition, Tag("voxel_grid")
    ],  # * 3D grid of resources: Collection[Collection[Collection[Resource]]]
    Annotated[
        PoolResourceTypeDefinition, Tag("pool")
    ],  # * collection of consumables with no structure: Collection[Consumable]
    Annotated[
        SlotTypeDefinition, Tag("slot")
    ],  # * slot for a single resource: Container[Resource]
]



"""
Event types for the MADSci system.
"""

import logging
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Literal, Optional

from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.base_types import BaseModel, PathLike, new_ulid_str
from madsci.common.types.lab_types import ManagerDefinition, ManagerType
from madsci.common.validators import ulid_validator
from pydantic.functional_validators import field_validator
from sqlmodel import Field


class Event(BaseModel):
    """An event in the MADSci system."""

    event_id: str = Field(
        title="Event ID",
        description="The ID of the event.",
        default_factory=new_ulid_str,
    )
    event_type: "EventType" = Field(
        title="Event Type",
        description="The type of the event.",
        default_factory=lambda: EventType.UNKNOWN,
    )
    log_level: "EventLogLevel" = Field(
        title="Event Log Level",
        description="The log level of the event. Defaults to NOTSET. See https://docs.python.org/3/library/logging.html#logging-levels",
        default_factory=lambda: EventLogLevel.INFO,
    )
    event_timestamp: datetime = Field(
        title="Event Timestamp",
        description="The timestamp of the event.",
        default_factory=datetime.now,
    )
    source: Optional[OwnershipInfo] = Field(
        title="Source",
        description="Information about the source of the event.",
        default=None,
    )
    event_data: Any = Field(
        title="Event Data",
        description="The data associated with the event.",
        default_factory=dict,
    )

    is_ulid = field_validator("event_id", mode="after")(ulid_validator)


class EventLogLevel(int, Enum):
    """The log level of an event."""

    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


class EventClientConfig(BaseModel):
    """Configuration for an Event Client."""

    name: Optional[str] = Field(
        title="Event Client Name",
        description="The name of the event client.",
        default=None,
    )
    event_server_url: Optional[str] = Field(
        title="Event Server URL",
        description="The URL of the event server.",
        default=None,
    )
    log_level: int = Field(
        title="Event Client Log Level",
        description="The log level of the event client.",
        default=EventLogLevel.INFO,
    )
    source: Optional[OwnershipInfo] = Field(
        title="Source",
        description="Information about the source of the event client.",
        default=None,
    )
    log_dir: PathLike = Field(
        title="Log Directory",
        description="The directory to store logs in.",
        default_factory=lambda: Path("~") / ".madsci" / "logs",
    )


class EventType(str, Enum):
    """The type of an event."""

    UNKNOWN = "unknown"
    LOG = "log"
    LOG_DEBUG = "log_debug"
    LOG_INFO = "log_info"
    LOG_WARNING = "log_warning"
    LOG_ERROR = "log_error"
    LOG_CRITICAL = "log_critical"
    TEST = "test"
    # *Lab Events
    LAB_CREATE = "lab_create"
    LAB_START = "lab_start"
    LAB_STOP = "lab_stop"
    # *Node Events
    NODE_CREATE = "node_create"
    NODE_START = "node_start"
    NODE_STOP = "node_stop"
    NODE_CONFIG_UPDATE = "node_config_update"
    NODE_STATUS_UPDATE = "node_status_update"
    NODE_ERROR = "node_error"
    # *Workcell Events
    WORKCELL_CREATE = "workcell_create"
    WORKCELL_START = "workcell_start"
    WORKCELL_STOP = "workcell_stop"
    WORKCELL_CONFIG_UPDATE = "workcell_config_update"
    WORKCELL_STATUS_UPDATE = "workcell_status_update"
    # *Workflow Events
    WORKFLOW_CREATE = "workflow_create"
    WORKFLOW_START = "workflow_start"
    WORKFLOW_COMPLETE = "workflow_complete"
    WORKFLOW_ABORT = "workflow_abort"
    # *Experiment Events
    EXPERIMENT_CREATE = "experiment_create"
    EXPERIMENT_START = "experiment_start"
    EXPERIMENT_COMPLETE = "experiment_complete"
    EXPERIMENT_FAILED = "experiment_failed"
    EXPERIMENT_CANCELLED = "experiment_stop"
    EXPERIMENT_PAUSE = "experiment_pause"
    EXPERIMENT_CONTINUED = "experiment_continued"
    # *Campaign Events
    CAMPAIGN_CREATE = "campaign_create"
    CAMPAIGN_START = "campaign_start"
    CAMPAIGN_COMPLETE = "campaign_complete"
    CAMPAIGN_ABORT = "campaign_abort"
    # *Action Events
    ACTION_STATUS_CHANGE = "action_status_change"

    @classmethod
    def _missing_(cls, value: str) -> "EventType":
        value = value.lower()
        for member in cls:
            if member.lower() == value:
                return member
        raise ValueError(f"Invalid ManagerTypes: {value}")


class EventManagerDefinition(ManagerDefinition):
    """Definition for a Squid Event Manager"""

    manager_type: Literal[ManagerType.EVENT_MANAGER] = Field(
        title="Manager Type",
        description="The type of the event manager",
        default=ManagerType.EVENT_MANAGER,
    )
    host: str = Field(
        default="127.0.0.1",
        title="Server Host",
        description="The hostname or IP address of the Event Manager server.",
    )
    port: int = Field(
        default=8001,
        title="Server Port",
        description="The port number of the Event Manager server.",
    )
    db_url: str = Field(
        default="mongodb://localhost:27017",
        title="Database URL",
        description="The URL of the database used by the Event Manager.",
    )
    event_client_config: "EventClientConfig" = Field(
        default_factory=lambda: EventClientConfig(),
        title="Event Client Configuration",
        description="The configuration for a MADSci event client. This is used by the event manager to log it's own events/logs. Note that the event_server_url is ignored.",
    )



"""Types for MADSci Conditions."""

from enum import Enum
from typing import Annotated, Literal, Optional, Union

from madsci.common.types.base_types import BaseModel
from madsci.common.types.resource_types import GridIndex, GridIndex2D, GridIndex3D
from pydantic import Discriminator
from sqlmodel.main import Field


class ConditionTypeEnum(str, Enum):
    """Types of conditional check for a step"""

    RESOURCE_PRESENT = "resource_present"
    NO_RESOURCE_PRESENT = "no_resource_present"

    @classmethod
    def _missing_(cls, value: str) -> "ConditionTypeEnum":
        """Convert the value to lowercase"""
        value = value.lower()
        for member in cls:
            if member.lower() == value:
                return member
        raise ValueError(f"Invalid ConditionType: {value}")


class Condition(BaseModel, extra="allow"):
    """A model for the conditions a step needs to be run"""


class ResourceInLocationCondition(Condition):
    """A condition that checks if a resource is present"""

    condition_type: Literal[ConditionTypeEnum.RESOURCE_PRESENT] = Field(
        title="Condition Type",
        description="The type of condition to check",
        const=ConditionTypeEnum.RESOURCE_PRESENT,
    )
    location: Optional[str] = Field(
        title="Location",
        description="The name or ID of the location to check for a resource in",
        default=None,
    )
    key: Optional[Union[str, int, GridIndex, GridIndex2D, GridIndex3D]] = Field(
        title="Key",
        description="The key to check in the location's container resource",
        default=None,
    )


class NoResourceInLocationCondition(Condition):
    """A condition that checks if a resource is present"""

    condition_type: Literal[ConditionTypeEnum.NO_RESOURCE_PRESENT] = Field(
        title="Condition Type",
        description="The type of condition to check",
        const=ConditionTypeEnum.NO_RESOURCE_PRESENT,
    )
    location: Optional[str] = Field(
        title="Location",
        description="The name or ID of the location to check for a resource in",
        default=None,
    )
    key: Optional[Union[str, int, GridIndex, GridIndex2D, GridIndex3D]] = Field(
        title="Key",
        description="The key to check in the location's container resource",
        default=None,
    )


Conditions = Annotated[
    Union[ResourceInLocationCondition, NoResourceInLocationCondition],
    Discriminator(
        discriminator="condition_type",
    ),
]



"""Location types for MADSci."""

from datetime import datetime
from typing import Optional

from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.base_types import BaseModel, new_ulid_str
from madsci.common.validators import ulid_validator
from pydantic import Field
from pydantic.functional_validators import field_validator
from pydantic.types import Json


class Location(BaseModel):
    """A location in the lab."""

    location_name: str = Field(
        title="Location Name",
        description="The name of the location.",
    )
    location_id: str = Field(
        title="Location ID",
        description="The ID of the location.",
        default_factory=new_ulid_str,
    )
    description: Optional[str] = Field(
        title="Location Description",
        description="A description of the location.",
        default=None,
    )
    lookup: dict[str, Json] = Field(
        title="Location Representation Map",
        description="A dictionary of different representations of the location. Allows creating an association between a specific key (like a node name or id) and a relevant representation of the location (like joint angles, a specific actuator, etc).",
        default=[],
    )
    resource_id: Optional[str] = Field(
        title="Resource ID",
        description="The resource ID linked to the location, typically a container.",
        default=None,
    )
    reservation: Optional["LocationReservation"] = Field(
        title="Location Reservation",
        description="The reservation for the location.",
        default=None,
    )

    is_ulid = field_validator("location_id")(ulid_validator)


class LocationReservation(BaseModel):
    """Reservation of a MADSci Location."""

    owned_by: OwnershipInfo = Field(
        title="Owned By",
        description="Who has ownership of the reservation.",
    )
    created: datetime = Field(
        title="Created Datetime",
        description="When the reservation was created.",
    )
    start: datetime = Field(
        title="Start Datetime",
        description="When the reservation starts.",
    )
    end: datetime = Field(
        title="End Datetime",
        description="When the reservation ends.",
    )

    def check(self, ownership: OwnershipInfo) -> bool:
        """Check if the reservation is 1.) active or not, and 2.) owned by the given ownership."""
        return not (
            not self.owned_by.check(ownership)
            and self.start <= datetime.now()
            and self.end >= datetime.now()
        )



"""Types for Admin Commands."""

from enum import Enum

from madsci.common.types.base_types import BaseModel, Error
from sqlmodel.main import Field


class AdminCommands(str, Enum):
    """Valid Admin Commands to send to a Node"""

    SAFETY_STOP = "safety_stop"
    RESET = "reset"
    PAUSE = "pause"
    RESUME = "resume"
    CANCEL = "cancel"
    SHUTDOWN = "shutdown"
    LOCK = "lock"
    UNLOCK = "unlock"


class AdminCommandResponse(BaseModel):
    """Response from an Admin Command"""

    success: bool = Field(
        title="Admin Command Success",
        description="Whether the admin command was successful.",
        default=True,
    )
    errors: list[Error] = Field(
        title="Admin Command Errors",
        description="A list of errors that occurred while executing the admin command.",
        default_factory=list,
    )
